/*
 * 
 * P4_16 for Tofino ASIC
 * Written by Jonatan Langlet for Direct Telemetry Access
 * Translator pipeline
 * 
 */
#include <core.p4>
#include <tna.p4>

#define ETHERTYPE_IPV4 0x0800
#define ETHERTYPE_PITCHER 0x1337

#define IPv4_PROTO_UDP 0x11   // UDP 在 IPv4 protocol 字段的对应编号 

#define CHECKSUM_CACHE_REGISTER_SIZE 65536 // Make sure this fits into telemetry_checksum_cache_index_t
// #define NUM_COLLECTORS 256  // Ensure this fits in server_id_t

#define DTA_PORT_NUMBER 40040
#define DTA_ACK_PORT_NUMBER 40044

/* 四种原语的操作码 */
#define DTA_OPCODE_KEYWRITE 0x01
#define DTA_OPCODE_APPEND 0x02
#define DTA_OPCODE_KEYINCREMENT 0x03
#define DTA_OPCODE_POSTCARDER 0x04

/* RoCEv2 的端口号 (4791) */
#define ROCEV2_UDP_PORT 4791 
// #define ROCEV2_UDP_PORT 5000  // Used to bypass RDMA processing (for debugging)

#define POSTCARDER_CACHE_SIZE 32768


#ifndef MAX_SUPPORTED_QPS
	#define MAX_SUPPORTED_QPS 256 // Maximum number of supported QPs. Specifies table and register sizes
	// #define MAX_SUPPORTED_QPS 65536 // Used when benchmarking tons of QPs
#endif

typedef bit<32> ipv4_address_t;

typedef bit<32> debug_t;
typedef bit<16> random_t;
typedef bit<16> telemetry_checksum_t;
typedef bit<16> telemetry_checksum_cache_index_t; //Make sure this is log(CHECKSUM_CACHE_REGISTER_SIZE)
typedef bit<8>  redundancy_entry_num_t;
typedef bit<32> telemetry_key_t;

typedef bit<8>  hop_num_t;
typedef bit<32> postcarder_data_t;
typedef bit<15> postcarder_cache_index_t; //can fit 32 768


// 更新 Key-Write 中 data 的大小: 
// 注意: 必须是 2 的幂，这样才能进行 slot->addr 的转换
// 1. Set this value (must be equal to power of 2, including checksum)
// 2. Update the keywrite_data_h to reflect data size changes
// 3. Update keywrite_slot_size_B in switch_cpu.py
// 4. Update the traffic generator to send correct packets
// 5. Update the collector struct to handle this size (struct keywriteEntry)

/* Key-Write RDMA 原语的有效负载 (data+checksum) 大小 (单位: 字节) */
/* 这里 checksum 的大小始终为 4 字节, data 的大小自行定义, 只要二者的和为 2 的幂次即可  */

#define KEYWRITE_RDMA_PAYLOAD_SIZE 8
//#define KEYWRITE_RDMA_PAYLOAD_SIZE 16  // 12B data (3x4B)
//#define KEYWRITE_RDMA_PAYLOAD_SIZE 32  // 28B data (7x4B)


typedef bit<32> telemetry_key_checksum_t;
typedef bit<32> keyval_data_t;             // data for Key-Write
typedef bit<64> keyincrement_counter_t;    // Counter in Key-Increment


// #define DO_NACK_TRACKING
#ifndef NUM_TRACKED_NACKS
	#define NUM_TRACKED_NACKS 65536
#endif


// 更新 Batch 的大小 (这块是 Append 原语中的部分): 
// 注意: 确保 Batch 中元素的数量是 2 的幂次
// 1. Update values here
// 2. Update in append rdma payload header (automatic)
// 3. Modify register handling in batch control block (automatic)
#ifndef APPEND_BATCH_SIZE
    /* APPEND_BATCH_SIZE 必须是 NUM_APPEND_ENTRIES_IN_REGISTERS + 1
	   同时还要确保它是 2 的幂次，否则无法处理环形缓冲区的滚动 */
	#define APPEND_BATCH_SIZE 4
	/* NUM_APPEND_ENTRIES_IN_REGISTERS 应该是 BatchSize-1(即 APPEND_BATCH_SIZE - 1)
	   这就是我们使用寄存器在管道内存储的数量(毕竟是以 Batch 为单位进行传输的, 因此需要存放先前的数据等待 Batch 满) */
	#define NUM_APPEND_ENTRIES_IN_REGISTERS 3
	/* 单个附加有效载荷(Append Payload)的大小乘以 Batch 中的元素数量, 表明总共需要附加的有效载荷大小 */
	#define APPEND_RDMA_PAYLOAD_SIZE 16

	// #define APPEND_BATCH_SIZE 16                  // Batch Size
	// #define NUM_APPEND_ENTRIES_IN_REGISTERS 15    // Batch Size - 1
	// #define APPEND_RDMA_PAYLOAD_SIZE 64           // Append Payload(default 4 bytes) * Batch Size
	
	//No batching
	//#define APPEND_BATCH_SIZE 1
	//#define NUM_APPEND_ENTRIES_IN_REGISTERS 0
	//#define APPEND_RDMA_PAYLOAD_SIZE 4
#endif

typedef bit<32> batch_entry_num_t;  // Used as entry-identifier for building append batches (8 likely enough, but 32 because compiler issues)
typedef bit<32> append_data_t;      // data for append (4 bytes)

typedef bit<32> rdma_write_len_t;
typedef bit<32> iCRC_t;
typedef bit<32> remote_key_t;
typedef bit<24> queue_pair_t;
typedef bit<24> psn_t;         // RoCEv2 中的数据包序列号 (Packet sequence number)

typedef bit<32> list_id_t;     // Append 原语中的 List IDs 

typedef bit<16> qp_reg_index_t; // 用于为每个 QP 存放其 PSN. 该字段是充当那个寄存器的索引号

typedef bit<32> memory_slot_t;      // 内存插槽(空隙). 由 Key-Write 和 Append 原语所共享, 出于某些原因, 它们都被限制在最大 32 bits
typedef bit<64> memory_address_t;   // 物理内存地址(共 2^64)

typedef bit<32> drop_counter_t;     // 用于在重新同步期间停止 QP 流量

// 在 QP 重新同步期间（允许系统刷新）丢弃的 DTA-RDMA 转换的总次数
#ifndef QP_RESYNC_PACKET_DROP_NUM
	#define QP_RESYNC_PACKET_DROP_NUM 100000   // 100K (proved stable in our tests)
#endif

//#define DISABLE_CONGESTION_HANDLING //Defining this one will drop RDMA congestion acks in Ingress, preventing PSN resync

typedef bit<8> dta_seqnum_t;

// 定义不同的数据包类型 (Normal 和 Mirror), 用于桥接报头中
typedef bit<8>  pkt_type_t;
const pkt_type_t PKT_TYPE_NORMAL = 1;
const pkt_type_t PKT_TYPE_MIRROR = 2;

// 定义不同的镜像数据包类型 (I2E 和 E2E)
typedef bit<3> mirror_type_t;
const mirror_type_t MIRROR_TYPE_I2E = 1;
const mirror_type_t MIRROR_TYPE_E2E = 2;

// 14 bytes
header ethernet_h
{
	bit<48> dstAddr;
	bit<48> srcAddr;
	bit<16> etherType;
}

// 20 bytes
header ipv4_h
{
	bit<4> version;
	bit<4> ihl;
	bit<6> dscp;
	bit<2> ecn;
	bit<16> totalLen;
	bit<16> identification;
	bit<3> flags;
	bit<13> fragOffset;
	bit<8> ttl;
	bit<8> protocol;
	bit<16> hdrChecksum;
	ipv4_address_t srcAddr;
	ipv4_address_t dstAddr;
}

// 8 bytes
header udp_h
{
	bit<16> srcPort;
	bit<16> dstPort;
	bit<16> length;
	bit<16> checksum;
}

// 2 bytes
header dta_ack_h
{
	dta_seqnum_t seqnum;
	bit<1> nack;
	bit<7> reserved;
}

/* DTA 基本报头(规定操作码和其他处理字段) */
header dta_base_h
{
	bit<8> opcode;
	#ifdef DO_NACK_TRACKING
		dta_seqnum_t seqnum;
		bit<1> immediate;        // CPU should be informed of this report
		bit<1> retransmitable;   // This report can be retransmitted, do seqnum tracking
		bit<6> reserved;
	#else
		bit<1> immediate;
		bit<7> reserved;
	#endif
	//bit<16> collectorID; // Do we need this field? Or maybe instead just use the IP?
}

/* 由 Key-Write 和 Key-Increment 所共享, 但是他们有着不同的子报头 */
header dta_keyval_static_h
{
	bit<8> redundancyLevel;
	telemetry_key_t key;         // 4 bytes
}
header dta_keyincrement_h
{
	keyincrement_counter_t counter;    // 8 bytes
}

// 这将在 DTA 和 RDMA Key-Write 报头之间共享，以简化转换过程
header keywrite_data_h
{
	keyval_data_t data1;         // 4 bytes
	/*keyval_data_t data2;
	keyval_data_t data3;
	keyval_data_t data4;
	keyval_data_t data5;
	keyval_data_t data6;
	keyval_data_t data7;*/
}

/* Append 原语的报头(由列表 ID 和 data 组成) */
header dta_append_static_h
{
	list_id_t listID;
	append_data_t data;     // 4 bytes
}

/* Postcarder 原语的报头(由 key, hop_num 和 data 组成) */
header dta_postcarder_h
{
	bit<32> key;
	hop_num_t hopNum;            // 1 bytes
	postcarder_data_t data;      // 4 bytes
}

// Base Transport Header (12 bytes)
header infiniband_bth_h
{
	bit<8> opcode;
	bit<1> solicitedEvent;
	bit<1> migReq;
	bit<2> padCount;
	bit<4> transportHeaderVersion;
	bit<16> partitionKey;
	bit<1> fRes;
	bit<1> bRes;
	bit<6> reserved1;
	bit<24> destinationQP;
	bit<1> ackRequest;
	bit<7> reserved2;
	psn_t packetSequenceNumber;
}

// RDMA Extended Transport Header (16 bytes)
// See Infiniband spec page 254
// RDMA 的 Read 或 Write 操作的第一条消息包含一个 RDMA ETH
// 它指定要 Read 或 Write 的数据缓冲区的虚拟地址、R_Key 和总长度
header infiniband_reth_h
{
	memory_address_t virtualAddress;
	bit<32> rKey;
	bit<32> dmaLength;
}

// See Infiniband spec page 242
header infiniband_atomiceth_h
{
	memory_address_t virtualAddress;
	bit<32> rKey;
	bit<64> data;
	bit<64> compare;
}

// 4+4 bytes(实际只有大小为 4 个字节的检验和字段, 数据字段被 keywrite_data_h 所替代了)
header rdma_payload_keyval_h
{
	telemetry_key_checksum_t checksum;      // 4 bytes
	//keyval_data_t data1;   // This is replaced by the shared keywrite_data_h
}

// Append 原语的有效载荷报头。其大小取决于 Batch 中元素的数量, 每个元素的大小为 4 字节
header rdma_payload_append_h
{
	append_data_t data1;
	#if APPEND_BATCH_SIZE >= 2
		append_data_t data2;
	#endif
	#if APPEND_BATCH_SIZE >= 4
		append_data_t data3;
		append_data_t data4;
	#endif
	#if APPEND_BATCH_SIZE >= 8
		append_data_t data5;
		append_data_t data6;
		append_data_t data7;
		append_data_t data8;
	#endif
	#if APPEND_BATCH_SIZE == 16
		append_data_t data9;
		append_data_t data10;
		append_data_t data11;
		append_data_t data12;
		append_data_t data13;
		append_data_t data14;
		append_data_t data15;
		append_data_t data16;
	#endif
}

// Postcarder 原语的有效载荷报头(这里默认最大为 5 跳, 每跳的数据大小为 4 字节)
header rdma_payload_postcarder_h
{
	bit<32> data1;
	bit<32> data2;
	bit<32> data3;
	bit<32> data4;
	bit<32> data5;
}

// 4 bytes
header infiniband_icrc_h
{
	bit<32> iCRC;
}


header mirror_h
{
	pkt_type_t pkt_type;
}

header mirror_bridged_metadata_h
{
	pkt_type_t pkt_type;
}

struct headers
{
	mirror_bridged_metadata_h bridged_md;
	ethernet_h ethernet;
	ipv4_h ipv4;
	udp_h udp;
	
	infiniband_bth_h bth;
	infiniband_reth_h reth;
	infiniband_atomiceth_h atomic_eth;
	
	#ifdef DO_NACK_TRACKING
		dta_ack_h dta_ack;
	#endif
	dta_base_h dta_base;
	dta_append_static_h dta_append;
	dta_postcarder_h dta_postcarder;
	rdma_payload_append_h rdma_payload_append;
	rdma_payload_postcarder_h rdma_payload_postcarder;
	
	
	dta_keyval_static_h dta_keyval;
	dta_keyincrement_h dta_keyincrement;
	
	rdma_payload_keyval_h rdma_payload_keyval;
	keywrite_data_h keywrite_data;
	
	infiniband_icrc_h icrc;
}

struct roce_icrc_fields_t
{
	/*
	infiniband_bth_h bth;
	infiniband_reth_h reth;
	infiniband_payload_h rdma_payload;
	*/
	
	// bth
	bit<8> bth_opcode;
	bit<1> bth_solicitedEvent;
	bit<1> bth_migReq;
	bit<2> bth_padCount;
	bit<4> bth_transportHeaderVersion;
	bit<16> bth_partitionKey;
	bit<1> bth_fRes;
	bit<1> bth_bRes;
	bit<6> bth_reserved1;
	bit<24> bth_destinationQP;
	bit<1> bth_ackRequest;
	bit<7> bth_reserved2;
	bit<24> bth_packetSequenceNumber;
	
	// reth
	memory_address_t reth_virtualAddress;
	bit<32> reth_rKey;
	bit<32> reth_dmaLength;
	
	// payload
	bit<32> payload_data1;
}


struct debug_digest_ingress_t
{
	debug_t debug;
	debug_t debug2;
}


struct ingress_metadata_t
{
	debug_t debug;      // Used for debug data (included in digest)
	debug_t debug2;
	random_t random;
	
	bit<1> send_debug_data;
	bit<1> generate_report;
	bit<1> detected_change;  // If a change was detected
	
	pkt_type_t pkt_type;
	MirrorId_t mirror_session;
	
	// telemetry_checksum_t telemetry_checksum;
	// keyval_data_t telemetry_data;
	// telemetry_key_t telemetry_key;
	telemetry_checksum_cache_index_t telemetry_checksum_cache_index;
	telemetry_checksum_t checksum_to_insert_into_cache;
	telemetry_checksum_t last_telemetry_checksum_cache_element;
	
	list_id_t list_reg_index;              // 这将赋给 Append 报头中的 listID
	batch_entry_num_t num_batched_items;   // 用于验证是否建立了一个 Append 的 Batch
	append_data_t append_data;
	
	bit<1> batch_ready;                    // 用于判断是否有一个可用的 Batch 
	
	#ifdef DO_NACK_TRACKING
		dta_seqnum_t prev_reporter_seqnum;
		bit<1> trigger_nack_response;
		ipv4_address_t ipAddr_tmp;
	#endif
}


struct egress_metadata_t
{
	bit<1> is_report_packet;               // 用于标识是否为报告数据包
	debug_t debug;
	
	telemetry_key_checksum_t telemetry_key_checksum;
	
	redundancy_entry_num_t redundancy_entry_num;
	
	roce_icrc_fields_t roce_icrc_fields;
	
	psn_t rdma_psn;
	remote_key_t remote_key;
	queue_pair_t queue_pair;
	memory_address_t memory_address_start;
	memory_slot_t collector_num_storage_slots;
	memory_slot_t destination_memory_slot;
	memory_address_t memory_write_offset;
	
	//bit<32> destination_memory_address_1;
	//bit<32> destination_memory_address_2;
	
	bit<16> rdma_payload_length; // 用于指示 RDMA 有效负载的长度/大小, 以字节为单位
	
	qp_reg_index_t qp_reg_index;
	
	
	// 用于 QP 的重新同步
	PortId_t egress_port;
	drop_counter_t drop_counter;
	bit<1> is_congestion_ack;
	bit<1> prevent_rdma_generation;   // 标记该数据包不生成 RDMA
	
	postcarder_cache_index_t postcarder_cache_index;
	// postcarder_data_t postcarder_towrite;
	postcarder_data_t postcarder_data1;
	postcarder_data_t postcarder_data2;
	postcarder_data_t postcarder_data3;
	postcarder_data_t postcarder_data4;
	postcarder_data_t postcarder_data5;
	bit<1> postcarder_ready_for_compile;
	bit<1> postcarder_collision;
	bit<32> stored_flowID;
	bit<8> cache_counter;
	
	bit<64> icrc_part_1;
	bit<64> icrc_part_2;
	bit<64> icrc_part_3;
	bit<64> icrc_part_4;
	bit<64> icrc_part_5;
	bit<64> icrc_part_6;
	bit<64> icrc_part_7;
	bit<64> icrc_part_8;
	bit<32> icrc_final;
}

parser TofinoIngressParser(packet_in pkt, inout ingress_metadata_t ig_md, out ingress_intrinsic_metadata_t ig_intr_md)
{
	state start
	{
		pkt.extract(ig_intr_md);
		transition select(ig_intr_md.resubmit_flag)
		{
			1 : parse_resubmit;
			0 : parse_port_metadata;
		}
	}

	state parse_resubmit
	{
		transition reject;
	}

	state parse_port_metadata
	{
		pkt.advance(64);  // Tofino 1
		transition accept;
	}
}

parser SwitchIngressParser(packet_in pkt, out headers hdr, out ingress_metadata_t ig_md, out ingress_intrinsic_metadata_t ig_intr_md)
{
	TofinoIngressParser() tofino_parser;
	
	state start 
	{
		tofino_parser.apply(pkt, ig_md, ig_intr_md);
		transition parse_ethernet;
	}
	
	state parse_ethernet
	{
		pkt.extract(hdr.ethernet);
		transition select(hdr.ethernet.etherType)
		{
			ETHERTYPE_IPV4: parse_ipv4;
			default: accept;
		}
	}
	
	state parse_ipv4
	{
		pkt.extract(hdr.ipv4);
		
		transition select(hdr.ipv4.protocol)
		{
			IPv4_PROTO_UDP: parse_udp;
			default: accept;
		}
	}
	
	state parse_udp
	{
		pkt.extract(hdr.udp);
		
		// 判断是 DTA 端口号还是 RoCEv2 端口号
		transition select(hdr.udp.dstPort)
		{
			DTA_PORT_NUMBER: parse_dta_base;
			ROCEV2_UDP_PORT: parse_rocev2_bth;
			default: accept;
		}
	}
	
	state parse_rocev2_bth
	{
		pkt.extract(hdr.bth);
		
		transition accept;
	}
	
	state parse_dta_base
	{
		pkt.extract(hdr.dta_base);
		
		// 判断是 DTA 中的四种原语的哪一个
		transition select(hdr.dta_base.opcode)
		{
			DTA_OPCODE_KEYWRITE: parse_dta_keyval;
			DTA_OPCODE_KEYINCREMENT: parse_dta_keyval;
			DTA_OPCODE_APPEND: parse_dta_append; 
			DTA_OPCODE_POSTCARDER: parse_dta_postcarder; 
			default: accept;
		}
	}
	
	// Key-Write 和 Key-Increment 两个原语共享该状态解析操作(都有着 dta_keyval 这个报头)
	state parse_dta_keyval
	{
		pkt.extract(hdr.dta_keyval);
		
		transition select(hdr.dta_base.opcode)
		{
			DTA_OPCODE_KEYWRITE: parse_keywrite_data;
			DTA_OPCODE_KEYINCREMENT: parse_keyincrement_counter; 
			default: reject; //Must be one of those, wth
		}
	}

	state parse_keywrite_data
	{
		pkt.extract(hdr.keywrite_data);
		
		transition accept;
	}

	state parse_keyincrement_counter
	{
		pkt.extract(hdr.dta_keyincrement);
		
		transition accept;
	}

	state parse_dta_append
	{
		pkt.extract(hdr.dta_append);
		
		transition accept;
	}

	state parse_dta_postcarder
	{
		pkt.extract(hdr.dta_postcarder);
		
		transition accept;
	}
}


control ControlAppendBatchHandling(inout headers hdr, inout ingress_metadata_t ig_md, inout ingress_intrinsic_metadata_for_deparser_t ig_intr_dprsr_md)
{
	/* 该寄存器用于记录和更新当前 Batch 中元素的数量 */
	Register<batch_entry_num_t, list_id_t>(MAX_SUPPORTED_QPS) reg_num_batched_elements;
	RegisterAction<batch_entry_num_t, list_id_t, batch_entry_num_t>(reg_num_batched_elements) get_num_batched = {
		void apply(inout batch_entry_num_t num_batched, out batch_entry_num_t output)
		{
			// 首先输出当前 Batch 中的元素数量
			output = num_batched;
			
			// 如果元素数量已经达到最大能够存放的数量, 则将计数值置 0
			if(num_batched == NUM_APPEND_ENTRIES_IN_REGISTERS)
				num_batched = 0;
			// 否则, 将计数值加 1 
			else
				num_batched = num_batched + 1;
		}
	};
	
	// 下面的操作对 Batch 中的所有元素都是一样的，因此创建一个共享的 define 即可
	// 返回旧的值 (如果此时为 rdma-ready). 它们都将包含相同的值 (即最后一个 append 元素的值)
	// 更新 Batch 的 data (如果是 Batch 生成器的话)
	
	/* 数据交换操作 (将 stored_data 的值输出给 output, 然后将自定义入口元数据的 append 数据赋给 stored_data) */
	#define APPEND_EXCHANGE_ACTION \
	void apply(inout append_data_t stored_data, out append_data_t output) \
	{\
		output = stored_data; \
		stored_data = ig_md.append_data; \
	}

	/* 数据存储操作 (将自定义入口元数据的 append 数据赋给 stored_data) */
	#define APPEND_STORE_ACTION \
	void apply(inout append_data_t stored_data, out append_data_t output) \
	{\
		stored_data = ig_md.append_data; \
	}

	/* 数据获取操作 (将 stored_data 的值输出给 output) */
	#define APPEND_GET_ACTION \
	void apply(inout append_data_t stored_data, out append_data_t output) \
	{\
		output = stored_data; \
	}
	

	// 定义用于构建 Append 的 Batch 的寄存器
	// 在存储和获取时重复使用动作，从而限制占用的空间
	#if APPEND_BATCH_SIZE >= 2
		Register<append_data_t, list_id_t>(MAX_SUPPORTED_QPS) reg_batch_1;
		RegisterAction<append_data_t, list_id_t, append_data_t>(reg_batch_1) exchange_batch_1 = { APPEND_EXCHANGE_ACTION };
	#endif
	#if APPEND_BATCH_SIZE >= 4
		Register<append_data_t, list_id_t>(MAX_SUPPORTED_QPS) reg_batch_2;
		Register<append_data_t, list_id_t>(MAX_SUPPORTED_QPS) reg_batch_3;
		RegisterAction<append_data_t, list_id_t, append_data_t>(reg_batch_2) exchange_batch_2 = { APPEND_EXCHANGE_ACTION };
		RegisterAction<append_data_t, list_id_t, append_data_t>(reg_batch_3) exchange_batch_3 = { APPEND_EXCHANGE_ACTION };
	#endif
	#if APPEND_BATCH_SIZE >= 8
		Register<append_data_t, list_id_t>(MAX_SUPPORTED_QPS) reg_batch_4;
		Register<append_data_t, list_id_t>(MAX_SUPPORTED_QPS) reg_batch_5;
		Register<append_data_t, list_id_t>(MAX_SUPPORTED_QPS) reg_batch_6;
		Register<append_data_t, list_id_t>(MAX_SUPPORTED_QPS) reg_batch_7;
		RegisterAction<append_data_t, list_id_t, append_data_t>(reg_batch_4) exchange_batch_4 = { APPEND_EXCHANGE_ACTION };
		RegisterAction<append_data_t, list_id_t, append_data_t>(reg_batch_5) exchange_batch_5 = { APPEND_EXCHANGE_ACTION };
		RegisterAction<append_data_t, list_id_t, append_data_t>(reg_batch_6) exchange_batch_6 = { APPEND_EXCHANGE_ACTION };
		RegisterAction<append_data_t, list_id_t, append_data_t>(reg_batch_7) exchange_batch_7 = { APPEND_EXCHANGE_ACTION };
	#endif
	#if APPEND_BATCH_SIZE == 16
		Register<append_data_t, list_id_t>(MAX_SUPPORTED_QPS) reg_batch_8;
		Register<append_data_t, list_id_t>(MAX_SUPPORTED_QPS) reg_batch_9;
		Register<append_data_t, list_id_t>(MAX_SUPPORTED_QPS) reg_batch_10;
		Register<append_data_t, list_id_t>(MAX_SUPPORTED_QPS) reg_batch_11;
		Register<append_data_t, list_id_t>(MAX_SUPPORTED_QPS) reg_batch_12;
		Register<append_data_t, list_id_t>(MAX_SUPPORTED_QPS) reg_batch_13;
		Register<append_data_t, list_id_t>(MAX_SUPPORTED_QPS) reg_batch_14;
		Register<append_data_t, list_id_t>(MAX_SUPPORTED_QPS) reg_batch_15;
		RegisterAction<append_data_t, list_id_t, append_data_t>(reg_batch_8) exchange_batch_8 = { APPEND_EXCHANGE_ACTION };
		RegisterAction<append_data_t, list_id_t, append_data_t>(reg_batch_9) exchange_batch_9 = { APPEND_EXCHANGE_ACTION };
		RegisterAction<append_data_t, list_id_t, append_data_t>(reg_batch_10) exchange_batch_10 = { APPEND_EXCHANGE_ACTION };
		RegisterAction<append_data_t, list_id_t, append_data_t>(reg_batch_11) exchange_batch_11 = { APPEND_EXCHANGE_ACTION };
		RegisterAction<append_data_t, list_id_t, append_data_t>(reg_batch_12) exchange_batch_12 = { APPEND_EXCHANGE_ACTION };
		RegisterAction<append_data_t, list_id_t, append_data_t>(reg_batch_13) exchange_batch_13 = { APPEND_EXCHANGE_ACTION };
		RegisterAction<append_data_t, list_id_t, append_data_t>(reg_batch_14) exchange_batch_14 = { APPEND_EXCHANGE_ACTION };
		RegisterAction<append_data_t, list_id_t, append_data_t>(reg_batch_15) exchange_batch_15 = { APPEND_EXCHANGE_ACTION };
	#endif
	
	apply
	{
		hdr.rdma_payload_append.setValid(); //The Append RDMA payload will follow the packet into egress (unless it is dropped of course)
		
		// ListID 用作寄存器中的索引, 而寄存器中存放的是该列表当前 Batch 的元素数量
		ig_md.list_reg_index = hdr.dta_append.listID;
		
		// 获取到目前为止, 该列表对应的 Batch 的 append 数据的数量
		ig_md.num_batched_items = get_num_batched.execute(ig_md.list_reg_index);
		
		// 如果与最大能够存放的数量一致, 则表明当前 Batch 已经构建完成
		if( ig_md.num_batched_items == NUM_APPEND_ENTRIES_IN_REGISTERS )
			ig_md.batch_ready = 1;          // TODO: 根据计数寄存器进行设置
		// 当前 Batch 尚未构建完成
		else 
		{
			ig_md.batch_ready = 0;
			//ig_md.batch_ready = 1; //Used while benchmarking append batching performance. Egress POV sees 16-fold increase in data (if max batching)
		}
		
		
		// 将遥测数据存放到 ig_md 的 append_data 字段中 (此处的 ++ 是级联运算符, 或者说拼接)
		ig_md.append_data = hdr.dta_append.data[7:0] ++
							hdr.dta_append.data[15:8] ++
							hdr.dta_append.data[23:16] ++
							hdr.dta_append.data[31:24];
		
		// 有条件地提取和处理每个条目（更新或检索，取决于由谁来做，但需要同时做这两件事）
		#if APPEND_BATCH_SIZE >= 2
			if( ig_md.num_batched_items == 0 || ig_md.batch_ready == 1 )
				hdr.rdma_payload_append.data1 = exchange_batch_1.execute(ig_md.list_reg_index);
		#endif
		#if APPEND_BATCH_SIZE >= 4
			if( ig_md.num_batched_items == 1 || ig_md.batch_ready == 1 )
				hdr.rdma_payload_append.data2 = exchange_batch_2.execute(ig_md.list_reg_index);
			if( ig_md.num_batched_items == 2 || ig_md.batch_ready == 1 )
				hdr.rdma_payload_append.data3 = exchange_batch_3.execute(ig_md.list_reg_index);
		#endif
		#if APPEND_BATCH_SIZE >= 8
			if( ig_md.num_batched_items == 3 || ig_md.batch_ready == 1 )
				hdr.rdma_payload_append.data4 = exchange_batch_4.execute(ig_md.list_reg_index);
			if( ig_md.num_batched_items == 4 || ig_md.batch_ready == 1 )
				hdr.rdma_payload_append.data5 = exchange_batch_5.execute(ig_md.list_reg_index);
			if( ig_md.num_batched_items == 5 || ig_md.batch_ready == 1 )
				hdr.rdma_payload_append.data6 = exchange_batch_6.execute(ig_md.list_reg_index);
			if( ig_md.num_batched_items == 6 || ig_md.batch_ready == 1 )
				hdr.rdma_payload_append.data7 = exchange_batch_7.execute(ig_md.list_reg_index);
		#endif
		#if APPEND_BATCH_SIZE == 16
			if( ig_md.num_batched_items == 7 || ig_md.batch_ready == 1 )
				hdr.rdma_payload_append.data8 = exchange_batch_8.execute(ig_md.list_reg_index);
			if( ig_md.num_batched_items == 8 || ig_md.batch_ready == 1 )
				hdr.rdma_payload_append.data9 = exchange_batch_9.execute(ig_md.list_reg_index);
			if( ig_md.num_batched_items == 9 || ig_md.batch_ready == 1 )
				hdr.rdma_payload_append.data10 = exchange_batch_10.execute(ig_md.list_reg_index);
			if( ig_md.num_batched_items == 10 || ig_md.batch_ready == 1 )
				hdr.rdma_payload_append.data11 = exchange_batch_11.execute(ig_md.list_reg_index);
			if( ig_md.num_batched_items == 11 || ig_md.batch_ready == 1 )
				hdr.rdma_payload_append.data12 = exchange_batch_12.execute(ig_md.list_reg_index);
			if( ig_md.num_batched_items == 12 || ig_md.batch_ready == 1 )
				hdr.rdma_payload_append.data13 = exchange_batch_13.execute(ig_md.list_reg_index);
			if( ig_md.num_batched_items == 13 || ig_md.batch_ready == 1 )
				hdr.rdma_payload_append.data14 = exchange_batch_14.execute(ig_md.list_reg_index);
			if( ig_md.num_batched_items == 14 || ig_md.batch_ready == 1 )
				hdr.rdma_payload_append.data15 = exchange_batch_15.execute(ig_md.list_reg_index);
		#endif
		
		// Batch 的最后一个元素无需储存到寄存器中，直接赋给 rdma_payload_append 的对应字段即可
		#if APPEND_BATCH_SIZE == 1
			hdr.rdma_payload_append.data1 = ig_md.append_data; 
		#endif
		#if APPEND_BATCH_SIZE == 2
			hdr.rdma_payload_append.data2 = ig_md.append_data; 
		#endif
		#if APPEND_BATCH_SIZE == 4
			hdr.rdma_payload_append.data4 = ig_md.append_data; 
		#endif
		#if APPEND_BATCH_SIZE == 8
			hdr.rdma_payload_append.data8 = ig_md.append_data; 
		#endif
		#if APPEND_BATCH_SIZE == 16
			hdr.rdma_payload_append.data16 = ig_md.append_data; 
		#endif
		
		
		// 如果 Append 原语的当前 Batch 尚未就绪(即 Batch 中的元素还没有填满)
		// 则将丢弃当前的 Append 操作(也就是丢弃当前的数据包，此时并不会生成 RDMA 数据包)
		if( ig_md.batch_ready == 0 )  
			ig_intr_dprsr_md.drop_ctl = 1;
		
	}
}

control ControlProcessDTAPacket(inout headers hdr, inout ingress_metadata_t ig_md, inout ingress_intrinsic_metadata_for_tm_t ig_intr_tm_md, inout ingress_intrinsic_metadata_for_deparser_t ig_intr_dprsr_md)
{
	ControlAppendBatchHandling() AppendBatchHandling;
	
	#ifdef DO_NACK_TRACKING
		Register<dta_seqnum_t, bit<16>>(NUM_TRACKED_NACKS,0) reg_nack_tracker;
		RegisterAction<dta_seqnum_t, bit<16>, dta_seqnum_t>(reg_nack_tracker) proc_nack_tracker = {
			void apply(inout dta_seqnum_t seqnum, out dta_seqnum_t output)
			{
				// 仅在报头中存放的值是正确时, 才会递增 (其实是对序列号进行递增)
				if( seqnum == hdr.dta_base.seqnum - 1 ) // The incoming sequence number is valid :)
					seqnum = seqnum + 1;
				
				output = seqnum;
			}
		};
	#endif
	
	// 此处的表和动作适用于 Key-Increment and Key-Write (这里的起名可能具有误导性)
	/*
	 * Ipv4 destination required for egress port
	 * redundancy level is required to know number of clones (therefore mcastgroup)
	 */
	action prep_MultiWrite(bit<16> mcast_grp)
	{
		ig_intr_tm_md.mcast_grp_a = mcast_grp;
		ig_md.debug = (debug_t)mcast_grp;
	}
	table tbl_Prep_KeyWrite
	{
		key = {
			hdr.ipv4.dstAddr: exact;
			hdr.dta_keyval.redundancyLevel: exact;
		}
		actions = {
			prep_MultiWrite;
			@defaultonly NoAction;
		}
		default_action = NoAction;
		size=1024;
	}
	
	#ifdef DO_NACK_TRACKING
	action craft_nack()
	{
		// 移除 DTA 报头
		hdr.dta_base.setInvalid();
		hdr.dta_keyval.setInvalid();
		hdr.keywrite_data.setInvalid();
		hdr.dta_append.setInvalid();
		
		// 添加 DTA 的 ACK 报头
		hdr.dta_ack.setValid();
		
		// 交换源和目的 IP 地址 (为了发送回 Reporter 交换机)
		ig_md.ipAddr_tmp = hdr.ipv4.srcAddr;
		hdr.ipv4.srcAddr = hdr.ipv4.dstAddr;
		hdr.ipv4.dstAddr = ig_md.ipAddr_tmp;
		
		// 修改 UDP 的目的端口号, 以标记这是一个 DTA 的 ACK 包
		hdr.udp.dstPort = DTA_ACK_PORT_NUMBER;

		// 发送先前的 ACK (期望的值 - 1) 作为响应
		hdr.dta_ack.seqnum = ig_md.prev_reporter_seqnum;
		// 表明这是一个 DTA 的 NACK
		hdr.dta_ack.nack = 1;
	}
	#endif
	
	apply
	{
		ig_md.debug = 0x69;
		
		// 为 Egress 处的 DTA 操作流程做准备
		// 判断是否为 DTA 的键值操作(Key-Write 或 Key-Increment)
		if( hdr.dta_base.opcode == DTA_OPCODE_KEYWRITE || hdr.dta_base.opcode == DTA_OPCODE_KEYINCREMENT )
		{
			tbl_Prep_KeyWrite.apply();
		}
		// 判断是否为 Append-to-List 操作
		else if( hdr.dta_base.opcode == DTA_OPCODE_APPEND )
		{
			//hdr.dta_append.listID = get_list_id.execute(0); //Temporary during num QP test, ensuring list round-robin
			AppendBatchHandling.apply(hdr, ig_md, ig_intr_dprsr_md);
		}
		
		#ifdef DO_NACK_TRACKING
			// 只对可重传数据包进行序列号(seqnum)跟踪
			if(hdr.dta_base.retransmitable == 1) 
			{
				// 这里的索引应该与每一个 Reporter 交换机所绑定
				// 此处由于仅有一个 Reporter 交换机，因此才用 0 作为索引
				ig_md.prev_reporter_seqnum = proc_nack_tracker.execute(0);

				// 如果输入的序列号与寄存器返回的序列号不同, 才发送 NACK 响应（如果没有数据缺口的话，二者应该相同）
				if( hdr.dta_base.seqnum != ig_md.prev_reporter_seqnum )
				{
					ig_md.trigger_nack_response = 1;
					// ig_intr_dprsr_md.drop_ctl = 1; // 目前只需删除无效序列号的报告（以获取跟踪器的编号）
					craft_nack();  // 携带一个 NACK
				}
			}
		#endif
	}
}


control SwitchIngress(inout headers hdr, inout ingress_metadata_t ig_md, in ingress_intrinsic_metadata_t ig_intr_md, in ingress_intrinsic_metadata_from_parser_t ig_intr_prsr_md, inout ingress_intrinsic_metadata_for_deparser_t ig_intr_dprsr_md, inout ingress_intrinsic_metadata_for_tm_t ig_intr_tm_md)
{
	ControlProcessDTAPacket() ProcessDTAPacket;
	

	action forward(PortId_t port)
	{
		ig_intr_tm_md.ucast_egress_port = port;
		hdr.ipv4.ttl = hdr.ipv4.ttl - 1;
	}
	action to_cpu()
	{
		ig_intr_tm_md.ucast_egress_port = 64;
		ig_md.debug = 666;
	}
	action drop()
	{
		ig_intr_dprsr_md.drop_ctl = 1;
	}
	/* 根据目的 IPv4 地址来决定对数据包进行普通转发, 转发到 CPU 或丢弃操作 */
	table tbl_forward
	{
		key = {
			hdr.ipv4.dstAddr: exact;
		}
		actions = {
			forward;
			to_cpu;
			drop;
		}
		default_action = to_cpu;
		size=1024;
	}
	
	apply
	{
		ig_md.debug = 0;
		ig_md.send_debug_data = 0;

		tbl_forward.apply();
		
		// 处理一个 DTA 数据包 (如果 dta_base 报头生效的话)
		if( hdr.dta_base.isValid() )
		{
			ProcessDTAPacket.apply(hdr, ig_md, ig_intr_tm_md, ig_intr_dprsr_md);
		}
		
		// Key-Write 和 Key-Increment 数据包将直接跳过 tbl_forward 表, 因为它们进行的是多播操作
		if( !hdr.dta_keyval.isValid() )
		{
			// 已注释 (用于测试)
			// tbl_forward.apply();
		}
			
		// 将 RDMA 的 Acks 反弹到与它们到达时相同的那个端口，以确保它们最终到达正确的出口管道
		if( hdr.bth.isValid() )
		{
			if(hdr.bth.opcode == 17 && hdr.ipv4.srcAddr == 0xc0a8fd81)
			{
				ig_intr_tm_md.ucast_egress_port = ig_intr_md.ingress_port;   // 反弹操作
				ig_md.send_debug_data = 1;
				ig_md.debug = (debug_t)hdr.bth.packetSequenceNumber;
				ig_md.debug2 = (debug_t)hdr.bth.destinationQP;
				#ifdef DISABLE_CONGESTION_HANDLING
					ig_intr_dprsr_md.drop_ctl = 1;
				#endif
			}
		}
		else
		{
			ig_md.debug = 0;
			ig_md.debug2 = 0;
		}
			
		// 为 Egress 准备桥接元数据
		hdr.bridged_md.setValid();
		// 镜像操作会覆写该字段
		hdr.bridged_md.pkt_type = PKT_TYPE_NORMAL; 
		
	}
}


control SwitchIngressDeparser(packet_out pkt, inout headers hdr, in ingress_metadata_t ig_md, in ingress_intrinsic_metadata_for_deparser_t ig_intr_dprsr_md)
{
	Digest<debug_digest_ingress_t>() debug_digest;
	Mirror() mirror;
	
	apply
	{
		// 此处是要执行 Digest 操作 (其实是为了发送 Debug 数据)
		if( ig_md.send_debug_data == 1 )
		{
			debug_digest.pack({
				ig_md.debug,
				ig_md.debug2
			});
		}
		
		// 镜像操作
		if (ig_intr_dprsr_md.mirror_type == MIRROR_TYPE_I2E)
		{
			// Emit Mirror，并附加上 mirror_h 报头
			mirror.emit<mirror_h>(ig_md.mirror_session, {ig_md.pkt_type});
		}
		
		pkt.emit(hdr);
	}
}


parser TofinoEgressParser(packet_in pkt, out egress_intrinsic_metadata_t eg_intr_md)
{
	state start
	{
		pkt.extract(eg_intr_md);
		transition accept;
	}
}


parser SwitchEgressParser(packet_in pkt, out headers hdr, out egress_metadata_t eg_md, out egress_intrinsic_metadata_t eg_intr_md)
{
	TofinoEgressParser() tofino_parser;

	state start
	{
		tofino_parser.apply(pkt, eg_intr_md);
		transition parse_metadata;
	}
	
	state parse_metadata
	{
		mirror_h mirror_md = pkt.lookahead<mirror_h>();
		
		// 根据镜像元数据中的 pkt_type 字段决定下一步要执行的解析状态
		transition select(mirror_md.pkt_type)
		{
			PKT_TYPE_MIRROR: parse_mirror_md;
			PKT_TYPE_NORMAL: parse_bridged_md;
			default: accept;
		}
	}
	
	state parse_bridged_md
	{
		pkt.extract(hdr.bridged_md);
		transition parse_ethernet;
	}
	
	state parse_mirror_md
	{
		// 如果是镜像数据包, 在本方案中表示是遥测报告数据包
		eg_md.is_report_packet = 1;
		
		mirror_h mirror_md;
		pkt.extract(mirror_md);
		transition parse_ethernet;
	}
	
	state parse_ethernet
	{
		pkt.extract(hdr.ethernet);
		transition select(hdr.ethernet.etherType)
		{
			ETHERTYPE_IPV4: parse_ipv4;
			default: accept;
		}
	}
	
	state parse_ipv4
	{
		pkt.extract(hdr.ipv4);
		
		transition select(hdr.ipv4.protocol)
		{
			IPv4_PROTO_UDP: parse_udp;
			default: accept;
		}
	}
	
	state parse_udp
	{
		pkt.extract(hdr.udp);
		
		// 判断是 DTA 数据包还是 RoCEv2 数据包
		transition select(hdr.udp.dstPort)
		{
			DTA_PORT_NUMBER: parse_dta_base;
			ROCEV2_UDP_PORT: parse_rocev2_bth;
			default: accept;
		}
	}
	
	state parse_rocev2_bth
	{
		pkt.extract(hdr.bth);
		
		// 将目的队列对 (destinationQP) 赋给自定义出口固有元数据的 queue_pair
		eg_md.queue_pair = hdr.bth.destinationQP;
		// 目前无需对 BTH 进行更深入的分析，因此直接 accept 即可
		transition accept; 
	}
	
	state parse_dta_base
	{
		pkt.extract(hdr.dta_base);
		
		transition select(hdr.dta_base.opcode)
		{
			DTA_OPCODE_KEYWRITE: parse_dta_keyval;
			DTA_OPCODE_KEYINCREMENT: parse_dta_keyval; 
			DTA_OPCODE_APPEND: parse_dta_append; 
			default: accept;
		}
	}
	
	state parse_dta_keyval
	{
		pkt.extract(hdr.dta_keyval);
		
		transition select(hdr.dta_base.opcode)
		{
			DTA_OPCODE_KEYWRITE: parse_keywrite_data;
			DTA_OPCODE_KEYINCREMENT: parse_keyincrement_counter; 
			default: reject; //Must be one of those, wth
		}
	}
	state parse_keywrite_data
	{
		pkt.extract(hdr.keywrite_data);
		
		transition accept;
	}
	state parse_keyincrement_counter
	{
		pkt.extract(hdr.dta_keyincrement);
		
		transition accept;
	}
	
	state parse_dta_append
	{
		pkt.extract(hdr.dta_append);
		pkt.extract(hdr.rdma_payload_append);
		
		transition accept;
	}
}


// 重新同步情况下的速率限制 (丢弃 X 个数据包)
// 以每个出口端口 (per-egress-port) 为基础 (因为我们需要对网卡进行速率限制)
// 现在，我们还将 NACK 跟踪放在此处
control ControlRDMARatelimit(inout headers hdr, inout egress_metadata_t eg_md)
{
	// 寄存器中存放着每个出口端口对应的还可丢弃的数据包数量
	Register<drop_counter_t, PortId_t>(1024,0) reg_rdma_drop_counter;

	// 用于对某个出口端口当前的计数值进行递减操作
	RegisterAction<drop_counter_t, PortId_t, drop_counter_t>(reg_rdma_drop_counter) get_drop_counter = {
		void apply(inout drop_counter_t counter, out drop_counter_t output)
		{
			// 对 counter 的值进行递减(每次只减 1)
			if( counter > 0 )
				counter = counter - 1;
			output = counter;
		}
	};
	// 用于初始化某个出口端口的计数值
	RegisterAction<drop_counter_t, PortId_t, drop_counter_t>(reg_rdma_drop_counter) initiate_drop_counting = {
		void apply(inout drop_counter_t counter, out drop_counter_t output)
		{
			counter = QP_RESYNC_PACKET_DROP_NUM;
			output = counter;
		}
	};
	
	// 用于设置某个队列对 (QP) 的索引号
	action set_qp_reg_num(qp_reg_index_t qp_reg_index)
	{
		eg_md.qp_reg_index = qp_reg_index;
	}
	table tbl_get_qp_reg_num
	{
		key = {
			eg_md.queue_pair: exact;
		}
		actions = {
			set_qp_reg_num;
			NoAction;
		}
		default_action = NoAction;
		size=1024;
	}
	
	apply
	{
		if(eg_md.is_congestion_ack == 1)
		{
			// 开始强制丢弃该出口端口的数据包(即先初始化丢包计数值)
			initiate_drop_counting.execute(eg_md.egress_port);
			// 重新同步 PSN 计数器时需要执行这步操作
			tbl_get_qp_reg_num.apply();
		}
		// 如果是 DTA 流量
		else if(hdr.dta_base.isValid()) 
		{
			// 获取该出口端口当前还可丢弃的数据包数量
			eg_md.drop_counter = get_drop_counter.execute(eg_md.egress_port);
			
			// 如果 drop_counter 大于 0 则应丢弃该数据包（忽略 RDMA 生成）
			if( eg_md.drop_counter > 0 )
			{
				// 这将绕过 PSN 增量器，并在 Deparser 中直接丢弃
				eg_md.prevent_rdma_generation = 1;
			}
		}
		
	}
}


// 假定 Append 原语的 RDMA 有效负载已在入口的 Batching 阶段准备就绪
control ControlPrepareAppend(inout headers hdr, inout egress_metadata_t eg_md)
{
	// 处理环形缓冲区中的 HEAD 指针(每个 QP 对应一个环形缓冲区)
	Register<memory_slot_t, qp_reg_index_t>(MAX_SUPPORTED_QPS) reg_head_pointer;
	// 用于获取环形缓冲区中 HEAD 指针的值, 并执行一些处理
	RegisterAction<memory_slot_t, qp_reg_index_t, memory_slot_t>(reg_head_pointer) get_head_offset = {
		void apply(inout memory_slot_t head, out memory_slot_t output)
		{
			// 如果当前 HEAD 指针的值小于收集器中可用的存储插槽的总数 (表明还是有可以存储的位置)
			if( head < eg_md.collector_num_storage_slots )
			{
				output = head;
				// 根据 Batch 中包含的元素数量来对 HEAD 指针进行移动
				head = head + APPEND_BATCH_SIZE;
			}
			else
			{
				output = 0;
				// 将 HEAD 指针的值重置为 0 (即环形缓冲区的起点)
				head = 0;
			}
			//This functionality disables HEAD wrap-around (used during PSN resync tests)
			//output = head;
			//head = head + APPEND_BATCH_SIZE;
		}
	};
	
	//TODO: re-merge tables. No need anymore to split them
	// 假设 dstIP 已经指向正确的列表收集器 (List Collector)

	// 该表用于根据 ListID 来获取并设置对应服务器的信息(remote_key, queue_pair, memory_address_start)
	action set_server_info_1(remote_key_t remote_key, queue_pair_t queue_pair, memory_address_t memory_address_start)
	{
		eg_md.remote_key = remote_key;
		eg_md.queue_pair = queue_pair;
		eg_md.memory_address_start = memory_address_start;
	}
	table tbl_getCollectorMetadataFromListID_1
	{
		key = {
			hdr.dta_append.listID: exact;
		}
		actions = {
			set_server_info_1;
		}
		// 单个 Translator 不可能负责比这更多的工作
		size = MAX_SUPPORTED_QPS;
	}

	// 该表用于根据 ListID 来获取并设置对应服务器的另外一些信息(collector_num_storage_slots, qp_reg_index)
	action set_server_info_2(memory_slot_t collector_num_storage_slots, qp_reg_index_t qp_reg_index)
	{
		eg_md.collector_num_storage_slots = collector_num_storage_slots;
		eg_md.qp_reg_index = qp_reg_index;
	}
	table tbl_getCollectorMetadataFromListID_2
	{
		key = {
			hdr.dta_append.listID: exact;
		}
		actions = {
			set_server_info_2;
		}
		// 单个 Translator 不可能负责比这更多的工作
		size = MAX_SUPPORTED_QPS;
	}
	
	apply
	{
		// 根据 ListID 获取对应 Collector 的信息
		tbl_getCollectorMetadataFromListID_1.apply();
		tbl_getCollectorMetadataFromListID_2.apply();
		
		eg_md.rdma_payload_length = APPEND_RDMA_PAYLOAD_SIZE; // 16Bytes, data * 4 (for batching)
		

		// 获取 HEAD 指针当前指向的插槽 (只有当 Batch 已满时，才会去更新 HEAD 指针)
		eg_md.destination_memory_slot = get_head_offset.execute(eg_md.qp_reg_index);
		// eg_md.destination_memory_slot = 0;   // 用于 RDMA 基线的测试 (仅写入环形缓冲区的第一个插槽)
		
		// 将内存插槽转换为内存偏移量, 即做一个乘 4 的操作 (因为每个插槽的大小为 4 Bytes)
		eg_md.memory_write_offset = (memory_address_t)(eg_md.destination_memory_slot);
		eg_md.memory_write_offset = eg_md.memory_write_offset * 4;
	}
}


control ControlPrepareKeyWrite(inout headers hdr, inout egress_metadata_t eg_md)
{
	Hash<memory_slot_t>(HashAlgorithm_t.CRC32) hash_memory_slot;
	
	// 这是假设冗余是连续处理的。因此，一个计数器就足够了。
	Register<redundancy_entry_num_t, bit<1>>(MAX_SUPPORTED_QPS) reg_redundancy_iterator;
	RegisterAction<redundancy_entry_num_t, bit<1>, redundancy_entry_num_t>(reg_redundancy_iterator) get_redundancy_number = {
		void apply(inout redundancy_entry_num_t stored, out redundancy_entry_num_t output)
		{
			// 首先将内部存储的 stored 输出给 output
			output = stored;
			
			// 如果当前存储的值大于 redundancyLevel - 1, 则需要将 stored 置 0
			if(stored >= hdr.dta_keyval.redundancyLevel - 1){
				stored = 0;
			}
			// 否则, 对 stored 进行递增
			else{
				stored = stored + 1;
			}
				
			
		}
	};
	
	// 每个 Collector 只支持一个 Key-Write 存储 (要想支持多个，请在数据包和此处包含一些 storage ID）
	// 此处为根据目的 IPv4 地址获取 Collector 服务器的信息
	action set_server_info(remote_key_t remote_key, queue_pair_t queue_pair, memory_address_t memory_address_start, memory_slot_t collector_num_storage_slots, qp_reg_index_t qp_reg_index)
	{
		eg_md.remote_key = remote_key;
		eg_md.queue_pair = queue_pair;
		eg_md.memory_address_start = memory_address_start;
		eg_md.collector_num_storage_slots = collector_num_storage_slots;
		eg_md.qp_reg_index = qp_reg_index;
	}
	table tbl_getCollectorMetadataFromIP
	{
		key = {
			hdr.ipv4.dstAddr: exact;
		}
		actions = {
			set_server_info;
		}
		// 单个 Translator 不可能负责比这更多的工作
		size = MAX_SUPPORTED_QPS;
	}
	
	
	/* 允许对 2 的幂次进行取模运算，将哈希到的内存插槽与可用的插槽进行绑定
	   我们假设 Collector 中最多有 2^32 个内存插槽 (可能更少，取决于其他内存计算组件的大小) */

	// 将目的内存插槽与实际可用的插槽进行绑定 (通过与 mask 进行按位与运算)
	action bound_memory_slot(memory_slot_t mask)
	{
		eg_md.destination_memory_slot = eg_md.destination_memory_slot & mask;
	}
	table tbl_bound_slot
	{
		key = {
			eg_md.collector_num_storage_slots: exact;
		}
		actions = {
			bound_memory_slot;
		}
		// 这里 Collecter 的内存插槽数量的最大值为 2^31, 因为存放 collector_num_storage_slots 的大小为 bit<32>
		const entries = {
			2: 				bound_memory_slot(0x00000001);
			4: 				bound_memory_slot(0x00000003);
			8: 				bound_memory_slot(0x00000007);
			16: 			bound_memory_slot(0x0000000f);
			32: 			bound_memory_slot(0x0000001f);
			64: 			bound_memory_slot(0x0000003f);
			128: 			bound_memory_slot(0x0000007f);
			256: 			bound_memory_slot(0x000000ff);
			512: 			bound_memory_slot(0x000001ff);
			1024: 			bound_memory_slot(0x000003ff);
			2048: 			bound_memory_slot(0x000007ff);
			4096: 			bound_memory_slot(0x00000fff);
			8192: 			bound_memory_slot(0x00001fff);
			16384: 			bound_memory_slot(0x00003fff);
			32768: 			bound_memory_slot(0x00007fff);
			65536: 			bound_memory_slot(0x0000ffff);
			131072: 		bound_memory_slot(0x0001ffff);
			262144: 		bound_memory_slot(0x0003ffff);
			524288: 		bound_memory_slot(0x0007ffff);
			1048576: 		bound_memory_slot(0x000fffff);
			2097152: 		bound_memory_slot(0x001fffff);
			4194304: 		bound_memory_slot(0x003fffff);
			8388608: 		bound_memory_slot(0x007fffff);
			16777216: 		bound_memory_slot(0x00ffffff);
			33554432: 		bound_memory_slot(0x01ffffff);
			67108864: 		bound_memory_slot(0x03ffffff);
			134217728: 		bound_memory_slot(0x07ffffff);
			268435456: 		bound_memory_slot(0x0fffffff);
			536870912: 		bound_memory_slot(0x1fffffff);
			1073741824: 	bound_memory_slot(0x3fffffff);
			2147483648: 	bound_memory_slot(0x7fffffff);
			//4294967296: 	bound_memory_slot(0xffffffff); //does not fit in 32-bit
			
		}
		size=64;
	}
	

	// 实际数据与解析后的 DTA 数据包保持一致，以简化处理过程
	action setRDMAPayload()
	{
		hdr.rdma_payload_keyval.setValid();
		
		eg_md.rdma_payload_length = KEYWRITE_RDMA_PAYLOAD_SIZE;  // 4 + 4 = 8 Bytes, checksum + data
		
		// 用校正的字节序方式来写入校验和 (网络一般是大端, 即 Big-Endian)
		hdr.rdma_payload_keyval.checksum = eg_md.telemetry_key_checksum[7:0] ++
									eg_md.telemetry_key_checksum[15:8] ++
									eg_md.telemetry_key_checksum[23:16] ++
									eg_md.telemetry_key_checksum[31:24];
	}
	
	apply
	{
		// 获取当前的 n 值
		eg_md.redundancy_entry_num = get_redundancy_number.execute(0);
		
		// 将 IP 与 Collector 的元数据作映射, 即通过目的 IP 地址获取对应 Collector 的元数据
		tbl_getCollectorMetadataFromIP.apply();
		
		// 设置 RDMA 的有效载荷
		setRDMAPayload();
		
		// 将 key 和 n (n in N) 哈希到一个内存地址中
		@stage(1)
		{
			// 哈希到内存插槽中 (不受内存大小的限制)
			// eg_md.destination_memory_slot = hash_memory_slot.get({hdr.dta_keywrite.key, eg_md.redundancy_entry_num});
			
			// 用修正后的字节序去哈希到内存插槽中 (以加速查询的速度)
			eg_md.destination_memory_slot = hash_memory_slot.get({hdr.dta_keyval.key[7:0] ++ hdr.dta_keyval.key[15:8] ++ hdr.dta_keyval.key[23:16] ++ hdr.dta_keyval.key[31:24], eg_md.redundancy_entry_num});
			
			// 根据内存大小来绑定内存插槽
			tbl_bound_slot.apply();
			// eg_md.destination_memory_slot = eg_md.destination_memory_slot % eg_md.collector_num_storage_slots;
			
			// 将内存插槽转换为物理内存地址的偏移量 (需要乘以有效载荷的字节数 (此处为 8，即向左位移 3))
			eg_md.memory_write_offset = (memory_address_t)(eg_md.destination_memory_slot);
			eg_md.memory_write_offset = eg_md.memory_write_offset * KEYWRITE_RDMA_PAYLOAD_SIZE;
			
		}
		
	}
}


control ControlPostcarder_cache(inout headers hdr, inout egress_metadata_t eg_md, inout postcarder_data_t md_result)(bit<32> checksum_seed, hop_num_t cached_hop)
{
	// 生成一个具有自定义 CRC 多项式的哈希函数( checksum_seed 由输入参数提供)
	CRCPolynomial<bit<32>>(checksum_seed, true, false, false, 32w0xFFFFFFFF, 32w0xFFFFFFFF) poly1;                               
	Hash<postcarder_data_t>(HashAlgorithm_t.CUSTOM, poly1) hash_flowID_checksum;
	
	Register<postcarder_data_t, postcarder_cache_index_t>(POSTCARDER_CACHE_SIZE, 0) reg_cache;
	/*RegisterAction<postcarder_data_t, postcarder_cache_index_t, bit<32>>(reg_cache) cache_write = {
		void apply(inout postcarder_data_t stored, out postcarder_data_t output)
		{
			stored = md_result; //Write into the cache
			output = 0; //We don't care about the output
		}
	};*/
	// 该动作用于提取已存储的值, 同时将新的值写入对应的位置
	RegisterAction<postcarder_data_t, postcarder_cache_index_t, bit<32>>(reg_cache) cache_replace = {
		void apply(inout postcarder_data_t stored, out postcarder_data_t output)
		{
			// 输出旧的值 (用于缓存碰撞时驱逐旧的 postcards)
			output = stored;
			// 将结果存放到寄存器的对应位置中
			stored = md_result;
		}
	};
	// 该动作仅用于提取已存储的值, 而不写入新的值
	RegisterAction<postcarder_data_t, postcarder_cache_index_t, bit<32>>(reg_cache) cache_extract = {
		void apply(inout postcarder_data_t stored, out postcarder_data_t output)
		{
			// 我们要提取这个值，而不将新值写入缓存中
			output = stored;
			// 因此需要把原有位置的值置为 0 
			stored = 0;
		}
	};
	
	apply
	{
		// 将输入内容存储在 md_result 中, 这个在最后将被替换
		// slot = data XOR h(flowID)
		// 将数据本身与对 key 的哈希结果进行异或, 得出 md_result
		md_result = hdr.dta_postcarder.data ^ hash_flowID_checksum.get({hdr.dta_postcarder.key});
		
		// 如果 DTA 的 Postcarder 报头中的当前跳数与 cached_hop 值相等, 则需要对其进行处理 (表明这是当前数据包对应的插槽)
		if( hdr.dta_postcarder.hopNum == cached_hop )
		{

			// 我们希望之后重置缓存，而 md_result 将仅保留其计算值
			if( eg_md.postcarder_ready_for_compile == 1 )
			{
				cache_extract.execute(eg_md.postcarder_cache_index); // 单纯地重置缓存
				// md_result = eg_md.postcarder_towrite; // 我们希望输出本应存储在缓存中的计算值
			}
			// 只需写入缓存 (要么是正常写入(忽略输出)，要么是碰撞写入(输出将被使用))
			else
			{
				md_result = cache_replace.execute(eg_md.postcarder_cache_index);
			}
		}
		// 如果我们要编译一个写入，也需要提取其他条目
		else if( eg_md.postcarder_ready_for_compile == 1 || eg_md.postcarder_collision == 1 ) 
		{
			md_result = cache_extract.execute(eg_md.postcarder_cache_index);
		}
		
	}
}


control ControlPreparePostcarder(inout headers hdr, inout egress_metadata_t eg_md)
{
	// 该寄存器用于存储 key 的值
	Register<bit<32>, postcarder_cache_index_t>(POSTCARDER_CACHE_SIZE,0) reg_cache_flowid;
	RegisterAction<bit<32>, postcarder_cache_index_t, bit<32>>(reg_cache_flowid) flowid_verify = {
		void apply(inout bit<32> stored, out bit<32> output)
		{
			// 如果没有发生碰撞，则输出 0；否则输出旧的键(即发生碰撞的话)
			if(stored == hdr.dta_postcarder.key){
				output = 0;
			}
			else{
				output = stored;
			}
			// 而且每次要更新 flowID (即 key 值)
			stored = hdr.dta_postcarder.key;
		}
	};
	
	// 该寄存器用于对缓存执行计数操作
	Register<bit<8>, postcarder_cache_index_t>(POSTCARDER_CACHE_SIZE,0) reg_cache_counter;
	// 递增后再输出
	RegisterAction<bit<8>, postcarder_cache_index_t, bit<8>>(reg_cache_counter) cache_counter_increment = {
		void apply(inout bit<8> stored, out bit<8> output)
		{
			stored = stored + 1;
			output = stored;
		}
	};
	// 计数器的重置
	RegisterAction<bit<8>, postcarder_cache_index_t, bit<8>>(reg_cache_counter) cache_counter_reset = {
		void apply(inout bit<8> stored, out bit<8> output)
		{
			stored = 1;
			output = 1;
		}
	};
	
	// 分别对第 1 跳到第 5 跳的缓存进行处理 (每个缓存都一个对应的自定义 CRC 多项式和 Control 控件实例)
	ControlPostcarder_cache(0x1e12a700, 1) cache_hop1;
	ControlPostcarder_cache(0x65b96595, 2) cache_hop2;
	ControlPostcarder_cache(0x49cf878b, 3) cache_hop3;
	ControlPostcarder_cache(0x36518f0d, 4) cache_hop4;
	ControlPostcarder_cache(0x7a40a908, 5) cache_hop5;
	

	Hash<postcarder_cache_index_t>(HashAlgorithm_t.CRC32) hash_cache_index;
	Hash<memory_slot_t>(HashAlgorithm_t.CRC32) hash_memory_slot;
	
	
	
	// 将目的内存插槽与实际可用的插槽进行绑定 (通过与 mask 进行按位与运算)(与 KeyVal 类似)
	action bound_memory_slot(memory_slot_t mask)
	{
		eg_md.destination_memory_slot = eg_md.destination_memory_slot & mask;
	}
	table tbl_bound_slot
	{
		key = {
			eg_md.collector_num_storage_slots: exact;
		}
		actions = {
			bound_memory_slot;
		}
		const entries = {
			2: 				bound_memory_slot(0x00000001);
			4: 				bound_memory_slot(0x00000003);
			8: 				bound_memory_slot(0x00000007);
			16: 			bound_memory_slot(0x0000000f);
			32: 			bound_memory_slot(0x0000001f);
			64: 			bound_memory_slot(0x0000003f);
			128: 			bound_memory_slot(0x0000007f);
			256: 			bound_memory_slot(0x000000ff);
			512: 			bound_memory_slot(0x000001ff);
			1024: 			bound_memory_slot(0x000003ff);
			2048: 			bound_memory_slot(0x000007ff);
			4096: 			bound_memory_slot(0x00000fff);
			8192: 			bound_memory_slot(0x00001fff);
			16384: 			bound_memory_slot(0x00003fff);
			32768: 			bound_memory_slot(0x00007fff);
			65536: 			bound_memory_slot(0x0000ffff);
			131072: 		bound_memory_slot(0x0001ffff);
			262144: 		bound_memory_slot(0x0003ffff);
			524288: 		bound_memory_slot(0x0007ffff);
			1048576: 		bound_memory_slot(0x000fffff);
			2097152: 		bound_memory_slot(0x001fffff);
			4194304: 		bound_memory_slot(0x003fffff);
			8388608: 		bound_memory_slot(0x007fffff);
			16777216: 		bound_memory_slot(0x00ffffff);
			33554432: 		bound_memory_slot(0x01ffffff);
			67108864: 		bound_memory_slot(0x03ffffff);
			134217728: 		bound_memory_slot(0x07ffffff);
			268435456: 		bound_memory_slot(0x0fffffff);
			536870912: 		bound_memory_slot(0x1fffffff);
			1073741824: 	bound_memory_slot(0x3fffffff);
			2147483648: 	bound_memory_slot(0x7fffffff);
			//4294967296: 	bound_memory_slot(0xffffffff); //does not fit in 32-bit
			
		}
		size=64;
	}
	
	// 此处为根据目的 IPv4 地址获取 Collector 服务器的信息
	action set_server_info(remote_key_t remote_key, queue_pair_t queue_pair, memory_address_t memory_address_start, memory_slot_t collector_num_storage_slots, qp_reg_index_t qp_reg_index)
	{
		eg_md.remote_key = remote_key;
		eg_md.queue_pair = queue_pair;
		eg_md.memory_address_start = memory_address_start;
		eg_md.collector_num_storage_slots = collector_num_storage_slots;
		eg_md.qp_reg_index = qp_reg_index;
	}
	table tbl_getCollectorMetadataFromIP
	{
		key = {
			hdr.ipv4.dstAddr: exact;
		}
		actions = {
			set_server_info;
		}
		size = MAX_SUPPORTED_QPS;
	}
	
	
	//TODOs:
	// 用 postcards 来制作实际的 RDMA
	// 对值进行编码 (通过哈希的方式)
	// 计算目的地址 (并且与之绑定)
	// 填充 getCollectorMetadataFromIP 表的表项/条目
	apply
	{
		// 通过对 dta_postcarder 报头中的 key 进行哈希来获取 postcarder 缓存的索引值
		eg_md.postcarder_cache_index = hash_cache_index.get({hdr.dta_postcarder.key});
		eg_md.rdma_payload_length = 20;   // 20 Bytes, data * 5 (for a full 5-hop path)
		
		// 根据目的 IP 地址获取 Collector 的元数据
		tbl_getCollectorMetadataFromIP.apply();
		
		// 检查该缓存中是否存在/发生 flowID 碰撞
		eg_md.stored_flowID = flowid_verify.execute(eg_md.postcarder_cache_index);
		// 如果输出的 flowID 为 0, 则表明发生了碰撞
		if( eg_md.stored_flowID != 0 )
		{
			eg_md.postcarder_collision = 1;    // 标记这里发生了碰撞
			cache_counter_reset.execute(eg_md.postcarder_cache_index);  // 重置 postcard 的计数器
			eg_md.cache_counter = 0;
		}
		// 如果没有发生碰撞的话
		else
		{
			// 对缓存的计数器中的值进行递增
			eg_md.cache_counter = cache_counter_increment.execute(eg_md.postcarder_cache_index);
			eg_md.postcarder_collision = 0;
		}
		
		// Currently hard-coded cache counter to 5 postcards
		// TODO: make this stated in postcards themselves?
		// 较短的数据会在从缓存中逐出时报告 (如果流量足够大的话，最终是会发生的)
		if(eg_md.cache_counter == 5)
		{
			eg_md.postcarder_ready_for_compile = 1; // 标记我们已经拿到所有的 postcards
		}
		else{
			eg_md.postcarder_ready_for_compile = 0; // 标记我们尚未拿到所有的 postcards
		}
			
		
		//Either extract cached entries, or write into one of these (depending on postcarder_ready_for_compile) 
		// 要么提取缓存的条目，要么写入其中之一 (取决于 postcarder_ready_for_compile)
		cache_hop1.apply(hdr, eg_md, eg_md.postcarder_data1);
		cache_hop2.apply(hdr, eg_md, eg_md.postcarder_data2);
		cache_hop3.apply(hdr, eg_md, eg_md.postcarder_data3);
		cache_hop4.apply(hdr, eg_md, eg_md.postcarder_data4);
		cache_hop5.apply(hdr, eg_md, eg_md.postcarder_data5);
		
		
		// 计算目的内存插槽
		eg_md.destination_memory_slot = hash_memory_slot.get({hdr.dta_postcarder.key});
		
		// 将内存插槽与内存大小绑定
		tbl_bound_slot.apply();
		
		// 将目的内存插槽转换为物理内存地址的偏移量
		eg_md.memory_write_offset = (memory_address_t)(eg_md.destination_memory_slot);
		// 由于需要乘以 2 的幂次，因此需要在 20Bytes 有效载荷的基础上增加 12Bytes 填充。因此，DRAM 中的插槽并不紧凑
		eg_md.memory_write_offset = eg_md.memory_write_offset * 32;
		
		
		// 如果我们应该精心设计 RDMA
		if( eg_md.postcarder_ready_for_compile == 1 || eg_md.postcarder_collision == 1 )
		{
			// 编译 RDMA 的有效载荷
			hdr.rdma_payload_postcarder.setValid();
			hdr.rdma_payload_postcarder.data1 = eg_md.postcarder_data1;
			hdr.rdma_payload_postcarder.data2 = eg_md.postcarder_data2;
			hdr.rdma_payload_postcarder.data3 = eg_md.postcarder_data3;
			hdr.rdma_payload_postcarder.data4 = eg_md.postcarder_data4;
			hdr.rdma_payload_postcarder.data5 = eg_md.postcarder_data5;
			eg_md.prevent_rdma_generation = 0;
		}
		else
		{
			eg_md.prevent_rdma_generation = 1;    // 标记我们不应该生成 RDMA
		}
	}
}


control ControlCraftRDMA(inout headers hdr, inout egress_metadata_t eg_md)
{
	// 分配 32 位的寄存器数组来保存 24 位的 PSN
	Register<bit<32>, qp_reg_index_t>(MAX_SUPPORTED_QPS) reg_rdma_sequence_number;
	RegisterAction<psn_t, qp_reg_index_t, psn_t>(reg_rdma_sequence_number) get_psn = {
		void apply(inout psn_t stored_psn, out psn_t output)
		{
			// 首先输出尚未递增的 PSN
			output = stored_psn;
			// 然后对 PSN 进行递增并覆盖原有的值
			stored_psn = stored_psn + 1;
		}
	};
	RegisterAction<psn_t, qp_reg_index_t, psn_t>(reg_rdma_sequence_number) set_psn = {
		void apply(inout psn_t stored_psn, out psn_t output)
		{
			// 将 PSN 重新同步为 ACK 获取的值
			stored_psn = eg_md.rdma_psn; 
			output = stored_psn;
		}
	};
	
	action setEthernet()
	{
		hdr.ethernet.setValid();

        // 此处已经修改
		hdr.ethernet.srcAddr = 0x6cec5a6239c8;   // Tofino CPU 的 MAC 地址
		hdr.ethernet.dstAddr = 0x08c0eb589289;   // Collector 的 MAC 地址
	}
	
	action setIP()
	{
		//TODO: finish filling out all fields with correct values
		hdr.ipv4.setValid();
		hdr.ipv4.ihl = 5;
		//DSCP field shall be set to the value in the Traffic Class component of the RDMA Address Vector associated with the packet.
		//hdr.ipv4.ecn = 0;
		hdr.ipv4.identification = 11381; //From random dumped packet
		//hdr.ipv4.ecn = 0b10; //According to dumped roce traffic
		hdr.ipv4.ecn = 0b00;
		//Total Length field shall be set to the length of the IPv4 packet in bytes including the IPv4 header and up to and including the ICRC.
		//hdr.ipv4.totalLen = 60+RDMA_PAYLOAD_SIZE_BYTES; //20+8+12+16+4 (+4 bytes payload) = 64
		hdr.ipv4.totalLen = ( \
			hdr.icrc.minSizeInBytes() + \
			hdr.reth.minSizeInBytes() + \
			hdr.bth.minSizeInBytes() + \
			hdr.udp.minSizeInBytes() + \
			hdr.ipv4.minSizeInBytes());
		//eg_md.rdma_payload_length + \
		
		hdr.ipv4.flags = 0b010;
		hdr.ipv4.fragOffset = 0;
		//Time to Live field shall be setto the value in the Hop Limit component of the RDMA Address Vector associated with the packet.
		hdr.ipv4.protocol = 0x11; //Set IPv4 proto to UDP
		//hdr.ipv4.dstAddr = eg_md.collector_ip; //Set address to collector address //dstAddr should already point to the collector (set at reporting switch)
		
		hdr.ipv4.srcAddr = 0xc0a8015b;   // 已经修改为 192.168.1.91
	}
	
	action setUDP()
	{
		//TODO: set the length value correctly
		hdr.udp.setValid();
		hdr.udp.srcPort = 10000; // Same as in initialization script
		hdr.udp.dstPort = ROCEV2_UDP_PORT;
		// RoCEv2 数据包的 UDP 报头中的 length 字段应设置为从 UDP 报头
		// 开始算起直到 ICRC 的第 4 个字节（包括 ICRC 的 4 个字节）的字节数
		// hdr.udp.length = 40+RDMA_PAYLOAD_SIZE_BYTES; //8 + 12+16+4+4
		hdr.udp.length = ( \
			hdr.icrc.minSizeInBytes() + \
			hdr.reth.minSizeInBytes() + \
			hdr.bth.minSizeInBytes() + \
			hdr.udp.minSizeInBytes());
		// eg_md.rdma_payload_length + \
		hdr.udp.checksum = 0; // UDP checksum SHOULD be 0
	}
	
	action setInfiniband_BTH()
	{
		//Infiniband Specification: https://cw.infinibandta.org/document/dl/7158 (page 239)
		//RoCE brief example: https://community.mellanox.com/s/article/rocev2-cnp-packet-format-example
		
		// TODO: fill the infiniband header fields
		hdr.bth.setValid();
		hdr.bth.opcode = 0x0a; // RDMA WRITE Only (is this correct operation?)
		hdr.bth.solicitedEvent = 0;
		hdr.bth.migReq = 1; // originally 0, set to 1 according to cloned packet
		hdr.bth.padCount = 0; // Set this depending of how many padded bytes at end of payload. (PL sent as multiple of 4 bytes)
		hdr.bth.transportHeaderVersion = 0; // Verify that this one is correct
		hdr.bth.partitionKey = 0xffff; // Identifies the partition that the desitnation QP or EE Context is a member (everyone always just uses 0xffff)
		hdr.bth.fRes = 0;
		hdr.bth.bRes = 0;
		hdr.bth.reserved1 = 0;
		hdr.bth.destinationQP = eg_md.queue_pair; // 指定目的地队列对 (QP) 标识符
		hdr.bth.ackRequest = 0; // Do we want an ACK? (default 0)
		hdr.bth.reserved2 = 0;
		// hdr.bth.packetSequenceNumber = eg_md.rdma_psn;
	}
	
	action setInfiniband_RETH()
	{
		//Infiniband Specification: https://cw.infinibandta.org/document/dl/7158 (page 239)
		//RoCE brief example: https://community.mellanox.com/s/article/rocev2-cnp-packet-format-example
		
		hdr.reth.setValid();
		
		// This is default. Possibly an issue in addition of 64b
		hdr.reth.virtualAddress = eg_md.memory_address_start + eg_md.memory_write_offset; 
		//Split addition into two steps. Unsure about order here. Can either handle overflow/carried bit?
		//hdr.reth.virtualAddress[63:32] = eg_md.memory_address_start[63:32] + eg_md.memory_write_offset[63:32];
		//hdr.reth.virtualAddress[31:0] = eg_md.memory_address_start[31:0] + eg_md.memory_write_offset[31:0];
		//eg_md.destination_memory_address_1 = eg_md.memory_address_start[63:32] + eg_md.memory_write_offset[63:32];
		//eg_md.destination_memory_address_2 = eg_md.memory_address_start[31:0] + eg_md.memory_write_offset[31:0];
		//hdr.reth.virtualAddress[63:32] = eg_md.destination_memory_address_1;
		//hdr.reth.virtualAddress[31:0] = eg_md.destination_memory_address_2;
		//hdr.reth.virtualAddress = eg_md.destination_memory_address_1 ++ eg_md.destination_memory_address_2;
		
		// TODO: change from cast to only modify LSB (saving staging cost)
		// Length, in bytes, of DMA operation. I guess size of memory to write? Payload size?
		hdr.reth.dmaLength = (bit<32>)eg_md.rdma_payload_length; 
	}
	
	// TODO: Add AETH generation support
	action setInfiniband_AETH()
	{
		hdr.atomic_eth.setValid();
		
		hdr.atomic_eth.virtualAddress = eg_md.memory_address_start + eg_md.memory_write_offset;
		hdr.atomic_eth.rKey = eg_md.remote_key;
		
		hdr.atomic_eth.data = hdr.dta_keyincrement.counter;
		//bit<64> data;
		//#bit<64> compare;
	}

	
	apply
	{
		//TODO: 计算 PSN 的索引号！并确保能够重新同步正确的索引号
		// eg_md.rdma_psn = (bit<24>)hdr.ipv4.srcAddr;  // Debugging
		
		// eg_md.rdma_psn = get_psn.execute(hdr.ipv4.dstAddr);  // 读取并更新该服务器(基于目标地址)的 RDMA 的 PSN
		setEthernet();
		@stage(2)
		{
		if( hdr.dta_base.isValid() ) // 将 DTA 转换为 RDMA
		{
		
			// 为 RoCEv2 制作报头
			setIP();
			hdr.ipv4.totalLen = hdr.ipv4.totalLen + eg_md.rdma_payload_length; // 这导致动作变大，无法适合单个阶段
			setUDP();
			hdr.udp.length = hdr.udp.length + eg_md.rdma_payload_length; // 与 IP 类似, 对于单个阶段来说也太大了
			setInfiniband_BTH();
			
			hdr.bth.packetSequenceNumber = get_psn.execute(eg_md.qp_reg_index);  // 读取并更新此 QP 的 RDMA PSN
			
			if( hdr.dta_keyincrement.isValid() ) // Key-Increment 使用 Fetch&Add 操作 (i.e., AETH)
			{
				hdr.bth.opcode = 0b00010100;     // Fetch&Add 的操作码
				setInfiniband_AETH();
			}
			// 其他操作, 如 Key-Write, Append, Postcarder 则使用 Write 操作(i.e., RETH)
			else       
			{
				setInfiniband_RETH();
				// 该字段被编译器错误地标记为 tagalong，因此移至此处以避免出现这种情况
				hdr.reth.rKey = eg_md.remote_key;
			}
			
			// setInfiniband_payload();
			
			hdr.icrc.setValid();
			
		}
		else if(eg_md.is_congestion_ack == 1)     // 该数据包应重新同步 PSN 计数器
		{
			eg_md.rdma_psn = hdr.bth.packetSequenceNumber;    // packetSequenceNumber 是我们将重置 PSN 的值
			set_psn.execute(eg_md.qp_reg_index);
		}
		}
		// 使 DTA 报头失效
		hdr.dta_base.setInvalid();
		hdr.dta_keyval.setInvalid();
		hdr.dta_keyincrement.setInvalid();
		hdr.dta_append.setInvalid();
		hdr.dta_postcarder.setInvalid();
	}
}




control SwitchEgress(inout headers hdr, inout egress_metadata_t eg_md, in egress_intrinsic_metadata_t eg_intr_md, in egress_intrinsic_metadata_from_parser_t eg_intr_from_prsr, inout egress_intrinsic_metadata_for_deparser_t eg_intr_md_for_dprsr, inout egress_intrinsic_metadata_for_output_port_t eg_intr_md_for_oport)
{
	Hash<telemetry_key_checksum_t>(HashAlgorithm_t.CRC32) hash_telemetry_key_checksum;
	
	ControlRDMARatelimit() RDMARatelimit;
	ControlPrepareKeyWrite() PrepareKeyWrite;
	ControlPrepareAppend() PrepareAppend;
	ControlPreparePostcarder() PreparePostcarder;
	ControlCraftRDMA() CraftRDMA;
	
	
	apply
	{
		// 检测这是否是来自服务器的拥塞 ACK
		// IP is hard-coded for now
		if(hdr.bth.opcode == 17 && hdr.ipv4.srcAddr == 0xc0a8fd81){
			eg_md.is_congestion_ack = 1;
		} 
		else
		{
			eg_md.is_congestion_ack = 0;
			eg_md.queue_pair = 0;
		}
		
		eg_md.egress_port = eg_intr_md.egress_port; // 稍后进行 RDMA 速率限制时需要使用此功能
		
		// 如果这是一个 DTA 操作
		if(hdr.dta_base.isValid()) 
		{
			// Key-Write 和 Key-Increment 原语
			if( hdr.dta_base.opcode == DTA_OPCODE_KEYWRITE || hdr.dta_base.opcode == DTA_OPCODE_KEYINCREMENT ) 
			{
				// 计算 key 校验和 (采用固定的字节序，以加快查询速度)
				eg_md.telemetry_key_checksum = hash_telemetry_key_checksum.get({hdr.dta_keyval.key[7:0] ++ hdr.dta_keyval.key[15:8] ++ hdr.dta_keyval.key[23:16] ++ hdr.dta_keyval.key[31:24]});
				// eg_md.telemetry_key_checksum = hash_telemetry_key_checksum.get({hdr.dta_keywrite.key});
				// 处理 DTA 的 Key-Write 操作
				PrepareKeyWrite.apply(hdr, eg_md);
			}
			// Append 原语
			else if( hdr.dta_base.opcode == DTA_OPCODE_APPEND ) 
			{
				PrepareAppend.apply(hdr, eg_md);
			}
			// Postcarder 原语
			else if( hdr.dta_base.opcode == DTA_OPCODE_POSTCARDER ) 
			{
				PreparePostcarder.apply(hdr, eg_md);
			}
		}
		

		// 对于 ACK 解析，请在此处添加 hdr.bth.isValid()（但请确保不会丢弃所有 RDMA 数据包）
		// DTA 数据包和拥塞 ACK 都会进入此处的代码块
		if( hdr.dta_base.isValid() || eg_md.is_congestion_ack == 1 )
		{
			RDMARatelimit.apply(hdr, eg_md);
			
			// 处理 RDMA 生成
			// 如果 RDMA 生成可以进行 (实际生成或 PSN 重新同步在此进行)
			if( eg_md.prevent_rdma_generation == 0 ){
				CraftRDMA.apply(hdr, eg_md);
			}
			// 不应产生 RDMA 的 DTA 流量应在逆解析器处丢弃
			else{
				eg_intr_md_for_dprsr.drop_ctl = 1; 
			}
		}
		
		// 这是一个拥塞 ACK，可用于重新同步 PSN。现在可以将其丢弃（无需发回 RDMA 网卡）
		if( eg_md.is_congestion_ack == 1 ){
			eg_intr_md_for_dprsr.drop_ctl = 1;
		}
		
		// 这是必需的，否则编译器会抛出随机错误
		eg_md.debug = (debug_t)eg_md.redundancy_entry_num;
		// eg_md.debug = (debug_t)eg_intr_md.egress_rid_first;
		// eg_md.debug = (debug_t)eg_intr_md.egress_rid_first;
		
		// 这永远不应该触发。除非我阻止调试被优化，否则编译器会失败......
		if( eg_md.debug == 1010 )
			hdr.ipv4.srcAddr = (bit<32>)eg_md.debug;
	}
} 

control SwitchEgressDeparser(packet_out pkt, inout headers hdr, in egress_metadata_t eg_md, in egress_intrinsic_metadata_for_deparser_t eg_dprsr_md)
{
	Checksum() ipv4_checksum;
	// Checksum<bit<32>>() icrc_checksum;
	apply
	{
		// Update IPv4 checksum
		hdr.ipv4.hdrChecksum = ipv4_checksum.update(
			{hdr.ipv4.version,
			 hdr.ipv4.ihl,
			 hdr.ipv4.dscp,
			 hdr.ipv4.ecn,
			 hdr.ipv4.totalLen,
			 hdr.ipv4.identification,
			 hdr.ipv4.flags,
			 hdr.ipv4.fragOffset,
			 hdr.ipv4.ttl,
			 hdr.ipv4.protocol,
			 hdr.ipv4.srcAddr,
			 hdr.ipv4.dstAddr});
			 
		pkt.emit(hdr.ethernet);
		pkt.emit(hdr.ipv4);
		pkt.emit(hdr.udp);
		// pkt.emit(hdr.dta_base);
		// pkt.emit(hdr.dta_keywrite);
		pkt.emit(hdr.bth);
		pkt.emit(hdr.reth);
		pkt.emit(hdr.atomic_eth);
		pkt.emit(hdr.rdma_payload_append);
		pkt.emit(hdr.rdma_payload_postcarder);
		pkt.emit(hdr.rdma_payload_keyval);
		pkt.emit(hdr.keywrite_data); //Send along the DTA keywrite data payload (if it is set valid)
		pkt.emit(hdr.icrc);
	}
}


Pipeline(SwitchIngressParser(),
	SwitchIngress(),
	SwitchIngressDeparser(),
	SwitchEgressParser(),
	SwitchEgress(),
	SwitchEgressDeparser()
) pipe;

Switch(pipe) main;
